import json
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Protocol

from ..cli.subprocess_utils import shell_output
from .title_parser import IssueTitle, parse_issue_title

if TYPE_CHECKING:
    from collections.abc import Sequence


@dataclass(frozen=True)
class Issue:
    entity_id: Optional[str]
    title: IssueTitle
    description: str


class IssueProvider(Protocol):
    def setup(self) -> None:
        """Any setup needed, including installing CLI tools, etc."""
        ...

    def get_latest_issues(self, in_progress_label: str) -> "Sequence[Issue]":
        ...

    def get_issues_assigned_to_me(self, in_progress_label: str) -> "Sequence[Issue]":
        ...

    def get_current_issue(self) -> Optional[Issue]:
        ...

    def label_issue(self, issue: Issue, label: str) -> None:
        ...

    def assign(self, issue: Issue, assignee: str) -> None:
        ...


class GithubIssueProvider(IssueProvider):
    def setup(self) -> None:
        pass

    def _values_to_issue(self, values: dict[str, str]) -> Issue:
        parsed_title = parse_issue_title(values["title"])
        return Issue(
            entity_id=str(values["number"]), title=parsed_title, description=values["body"]
        )

    def get_latest_issues(self, in_progress_label: str) -> "Sequence[Issue]":
        latest_issues = shell_output(
            f"gh issue list --limit 10 --json number,title,body --search 'is:open -label:{in_progress_label}'"
        )

        if latest_issues is None:
            return []

        return self._parse_github_json_str(latest_issues)

    def get_issues_assigned_to_me(self, in_progress_label: str) -> "Sequence[Issue]":
        """Get issues assigned to current user on current repo"""
        my_issues_cmd = shell_output(
            f"gh issue list --assignee @me  --search '-label:{in_progress_label}' --json number,title,body"
        )

        if my_issues_cmd is None:
            return []

        return self._parse_github_json_str(my_issues_cmd)

    def _parse_github_json_str(self, issue_str: str) -> "Sequence[Issue]":
        values = json.loads(issue_str)
        parsed_output = [self._values_to_issue(v) for v in values]
        return parsed_output

    def _create_label(self, label: str) -> None:
        shell_output(f"gh label create {label}")

    def _add_label_to_issue(self, issue: Issue, label: str) -> None:
        shell_output(f'gh issue edit "{issue.entity_id}" --add-label "{label}"')

    def get_current_issue(self) -> Optional[Issue]:
        current_branch: str = shell_output("git rev-parse --abbrev-ref HEAD")  # type: ignore

        branch_items = current_branch.split("/")

        if len(branch_items) != 3:
            return None

        return Issue(
            entity_id=branch_items[1],
            title=IssueTitle(prefix=branch_items[0], content=branch_items[2]),
            description="",
        )

    def label_issue(self, issue: Issue, label: str) -> None:
        if not issue.entity_id:
            return
        try:
            self._add_label_to_issue(issue, label)
        except Exception:
            try:
                self._create_label(label)
                self._add_label_to_issue(issue, label)
            except Exception as e:
                raise RuntimeError(f"Error labeling issue {issue.entity_id} with {label}") from e

    def assign(self, issue: Issue, assignee: str) -> None:
        if issue.entity_id:
            shell_output(f"gh issue edit {int(issue.entity_id)} --add-assignee {assignee}")



import enum
from typing import Annotated

import typer


class Location(str, enum.Enum):
    up = "up"
    top = "top"
    to = "to"
    down = "down"
    do = "do"
    bottom = "bottom"
    bo = "bo"
    trunk = "trunk"
    tr = "tr"

    @property
    def to_full_location(self) -> "Location":
        if self == Location.to:
            return Location.top
        if self == Location.do:
            return Location.down
        if self == Location.bo:
            return Location.bottom
        if self == Location.tr:
            return Location.trunk
        return self


LocationCLIOption = Annotated[Location, typer.Argument()]



from dataclasses import dataclass
from typing import TYPE_CHECKING, Protocol, Union

import typer
from iterfzf import iterfzf  # type: ignore
from rich.console import Console

from .provider import Issue
from .title_parser import parse_issue_title

if TYPE_CHECKING:
    from collections.abc import Sequence

console = Console()


class IssueSelecter(Protocol):
    ten_latest_prompt: str
    manual_prompt: str
    refresh_prompt: str

    def select_issue_dialog(self, issues: "Sequence[Issue]") -> Union[Issue, str]:
        ...


@dataclass(frozen=True)
class DefaultIssueSelecter(IssueSelecter):
    ten_latest_prompt: str = "Recent"
    manual_prompt: str = "Manual"
    refresh_prompt: str = "Refresh..."

    def _show_entry_dialog(self) -> str:
        return typer.prompt("Title")

    def _show_selection_dialog(self, issues: "Sequence[Issue]") -> str:
        issue_titles = [f"{issue.title.content} #{issue.entity_id}" for issue in issues]
        selection: str = iterfzf(
            [self.manual_prompt, *issue_titles, self.refresh_prompt, self.ten_latest_prompt]
        )  # type: ignore
        return selection

    def select_issue_dialog(self, issues: "Sequence[Issue]") -> Union[Issue, str]:
        selected_issue_title = self._show_selection_dialog(issues=issues)

        if selected_issue_title in [self.refresh_prompt, self.ten_latest_prompt]:
            return selected_issue_title

        if selected_issue_title == self.manual_prompt:
            selected_issue_title = self._show_entry_dialog()
            parsed_title = parse_issue_title(selected_issue_title)
            return Issue(entity_id=None, title=parsed_title, description="")

        return next(issue for issue in issues if issue.title.content in selected_issue_title)
